/*
 *  Sshtools - Java SSH2 API
 *
 *  Copyright (C) 2002 Lee David Painter.
 *
 *  Written by: 2002 Lee David Painter <lee@sshtools.com>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public License
 *  as published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package com.sshtools.j2ssh.transport.publickey;

import com.sshtools.j2ssh.io.ByteArrayReader;

import com.sshtools.j2ssh.transport.AlgorithmNotSupportedException;

import org.apache.log4j.Logger;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

import java.util.Iterator;


public class SshPrivateKeyFile {
    private static Logger log = Logger.getLogger(SshPrivateKeyFile.class);
    private SshPrivateKeyFormat format;
    private byte[] keyblob;

    protected SshPrivateKeyFile(byte[] keyblob, SshPrivateKeyFormat format) {
        this.keyblob = keyblob;
        this.format = format;
    }

    public byte[] getBytes() {
        return keyblob;
    }

    public byte[] getKeyBlob(String passphrase) throws InvalidSshKeyException {
        return format.decryptKeyblob(keyblob, passphrase);
    }

    public void changePassphrase(String oldPassphrase,
        String newPassphrase) throws InvalidSshKeyException {

        byte[] raw =format.decryptKeyblob(keyblob, oldPassphrase);
        keyblob = format.encryptKeyblob(raw, newPassphrase);

    }


    public static SshPrivateKeyFile parse(byte[] formattedKey)
        throws InvalidSshKeyException {
        log.info("Parsing private key file");

        // Try the default private key format
        SshPrivateKeyFormat format;

        format = SshPrivateKeyFormatFactory.newInstance(SshPrivateKeyFormatFactory
                .getDefaultFormatType());

        boolean valid = format.isFormatted(formattedKey);

        if (!valid) {
            log.info(
                "Private key is not in the default format, attempting parse with other supported formats");

            Iterator it = SshPrivateKeyFormatFactory.getSupportedFormats()
                                                    .iterator();
            String ft;

            while (it.hasNext() && !valid) {
                ft = (String) it.next();
                log.debug("Attempting " + ft);
                format = SshPrivateKeyFormatFactory.newInstance(ft);
                valid = format.isFormatted(formattedKey);
            }
        }

        if (valid) {
            return new SshPrivateKeyFile(formattedKey, format);
        } else {
            throw new InvalidSshKeyException(
                "The key format is not a supported format");
        }
    }

    public static SshPrivateKeyFile parse(File keyfile)
        throws InvalidSshKeyException, IOException {
        FileInputStream in = new FileInputStream(keyfile);
        byte[] data = new byte[in.available()];
        in.read(data);
        in.close();

        return parse(data);
    }

    public boolean isPassphraseProtected() {
        return format.isPassphraseProtected(keyblob);
    }

    /*public void changePassphrase(String oldPassphrase, String newPassphrase)
        throws InvalidSshKeyException {
        keyblob = format.changePassphrase(keyblob, oldPassphrase, newPassphrase);
    }*/

    public static SshPrivateKeyFile create(SshPrivateKey key,
        String passphrase, SshPrivateKeyFormat format)
        throws InvalidSshKeyException {
        byte[] keyblob = format.encryptKeyblob(key.getEncoded(), passphrase);

        return new SshPrivateKeyFile(keyblob, format);
    }

    public void setFormat(SshPrivateKeyFormat newFormat, String passphrase)
        throws InvalidSshKeyException {
        byte[] raw = this.format.decryptKeyblob(keyblob, passphrase);
        format = newFormat;
        keyblob = format.encryptKeyblob(raw, passphrase);
    }

    public SshPrivateKeyFormat getFormat() {
        return format;
    }

    public SshPrivateKey toPrivateKey(String passphrase)
        throws InvalidSshKeyException {
        try {
            byte[] raw = format.decryptKeyblob(keyblob, passphrase);
            SshKeyPair pair = SshKeyPairFactory.newInstance(getAlgorithm(raw));

            return pair.decodePrivateKey(raw);
        } catch (AlgorithmNotSupportedException anse) {
            throw new InvalidSshKeyException(
                "The public key algorithm for this private key is not supported");
        }
    }

    public String toString() {
        return new String(keyblob);
    }

    private String getAlgorithm(byte[] raw) {
        return ByteArrayReader.readString(raw, 0);
    }
}
